/****************************************************************************
*                                                                           *
*                            Third Year Project                             *
*                                                                           *
*                            An IBM PC Emulator                             *
*                          For Unix and X Windows                           *
*                                                                           *
*                             By David Hedley                               *
*                   Driver for external console (c) Sasha Chukov, 2000      *
*                                                                           *
* This program is Copyrighted.  Consult the file COPYRIGHT for more details *
*                                                                           *
****************************************************************************/
#include "global.h"
#include <stdio.h>
#include <termio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <fcntl.h>
#include <stdlib.h>
#include "icon.h"
#include "xstuff.h"
#include "vgahard.h"
#include "hardware.h"

#define	COPY		1
#define	DRAW_CHAR	2
#define	DRAW_LINE	3
#define	CLEAR_SCREEN	4
#define	WINDOW_SIZE	5
#define PUT_CURSOR	6
#define UNPUT_CURSOR	7
#define NEW_CURSOR	8
#define READ_KEYCODE	10

#define MYDELAY		60

extern BYTE scan_shift[];
extern BYTE scan_unshifted[];
extern BYTE scan_ctrl[];

int pyldin;

int oldCX=90;
int oldCY=90;

BYTE oldkey = 0xff;
BYTE oldSkey= 0xff;
BYTE oldMkey= 0xff;
void get_keycode();

void raw( void )
{
    struct termios it;

    if (tcgetattr(pyldin,&it)<0) err( "get attr" );
    it.c_lflag &= 0; //~(ICANON|ISIG|ECHO);
    it.c_iflag &= 0; // ~(INPCK|ISTRIP|IXON);
//it.c_iflag |= IGNPAR;
    it.c_oflag &=0; // ~(OPOST);
    it.c_cc[VMIN] = 1;
    it.c_cc[VTIME] = 0;

    if (cfsetospeed( &it, B19200 )<0) err( "set o speed" );
    if (cfsetispeed( &it, B19200 )<0) err( "set i speed" );
    if (tcsetattr(pyldin,TCSANOW,&it)<0) err( "set attr" );
}

void send_byte(char cmd)
{
    write(pyldin, &cmd, 1);
//    usleep(MYDELAY);
}

void send_word(int cmd)
{
    write(pyldin, (char *) &cmd+1, 1);
//    usleep(MYDELAY);
    write(pyldin, (char *) &cmd, 1);
//    usleep(MYDELAY);
}

BYTE recv_byte()
{
    char c;
    read(pyldin, &c, 1);
    return c;
}

void start_X(void)
{
    printf("Open COM2: channel ... ");
    pyldin = open("/dev/cua1", O_RDWR);
    if (!pyldin) {
	printf("Failed!\n");
	exit(-1);
    } else printf("Ok!\n");
    raw();
}

void end_X(void)
{
    printf("Close CON: channel\n");
    close(pyldin);
}

void process_Xevents(void)
{
//    printf("Xevent\n");
    get_keycode();
}

void flush_X(void)
{
//    printf("flush\n");
}

void copy(unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned nx, unsigned ny)
{
//    printf("copy %d, %d - %d, %d -> %d, %d\n", x1, y1, x2, y2, nx, ny);
    send_byte(COPY);
    recv_byte();	// wait request for data packet
    send_byte(x1);
    send_byte(y1);
    send_byte(x2);
    send_byte(y2);
    send_byte(nx);
    send_byte(ny);
    recv_byte();
}

BYTE pc2pyl(BYTE c)
{
    switch(c) {
	case 0xb0: return 0xd0;
	case 0xb1: return 0xd1;
	case 0xb2: return 0xd2;
	case 0xb3: return 0xd3;
	case 0xb4: return 0xd4;
	case 0xb9: return 0xc6;
	case 0xba: return 0xc7;
	case 0xbb: return 0xd7;
	case 0xbc: return 0xd8;
	case 0xbf: return 0xcf;
    }
    if ( (c & 0xf0) == 0xb0) return c+0x20;
//    if ( (c & 0xf0) == 0xd0) return c-0x20;
    if ( (c & 0xf0) == 0xe0) return c-0x30;
    return c;
}

void draw_char(unsigned x, unsigned y, char c, BYTE a)
{
    printf("draw_char %d, %d, %c, atr=%d\n", x, y, c, a);
    send_byte(DRAW_CHAR);
    recv_byte();	// wait request for data packet
    send_byte(x);
    send_byte(y);
    if (a != 0x70) send_byte(pc2pyl(c));
    else send_byte(0xfb);
    recv_byte();
}

void draw_line(unsigned x, unsigned y, char *s, unsigned l, BYTE a)
{
    int i;
    printf("draw_line %d, %d, %d, %s, atr=%02x\n", x, y, l, s, a);
    send_byte(DRAW_LINE);
    recv_byte();	// wait request for data packet
    send_byte(x);
    send_byte(y);
    send_word(l);
    for (i=0; i<l; i++)
        if (a==0x70 && i==1) { send_byte(0xfb); s++; }
	else send_byte(pc2pyl(*s++));
    recv_byte();
}

void clear_screen(void)
{
    char c;
    printf("clear_screen\n");
    send_byte(CLEAR_SCREEN);
    recv_byte();
}

void window_size(unsigned x, unsigned y)
{
    printf("window_size %d, %d\n", x, y);

    send_byte(WINDOW_SIZE);
    recv_byte();	// wait request for data packet
    send_byte(x);
    send_byte(y);
    recv_byte();

}

void new_cursor(int x, int y)
{
    printf("new_cursor %d, %d\n", x, y);
    send_byte(NEW_CURSOR);
    recv_byte();	// wait request for data packet
    send_byte(x>>1);
    send_byte(y>>1);
    recv_byte();
}

void put_cursor(unsigned x, unsigned y)
{
    if (oldCX != x || oldCY != y) {
        send_byte(PUT_CURSOR);
        recv_byte();	// wait request for data packet
        send_byte(x);
        send_byte(y);
        recv_byte();
        oldCX=x;
        oldCY=y;
        printf("put_cursor %d, %d\n", x, y);
    }
}

void unput_cursor(void)
{
//    printf("unput_cursor\n");
//    send_byte(UNPUT_CURSOR);
//    recv_byte();
}

DWORD asc2scan(BYTE asc)
{
    int i;
    
    switch(asc){
	case 0xc9:
	case 0xd5:
	case 0xe1: return 59 | 0xff00; // F1
	case 0xca:
	case 0xd6:
	case 0xe2: return 60 | 0xff00; // F2
	case 0xcb:
	case 0xd7:
	case 0xe3: return 61 | 0xff00; // F3
	case 0xcc:
	case 0xd8:
	case 0xe4: return 62 | 0xff00; // F4
	case 0xcd:
	case 0xd9:
	case 0xe5: return 63 | 0xff00; // F5
	case 0xce:
	case 0xda:
	case 0xe6: return 64 | 0xff00; // F6
	case 0xcf:
	case 0xdb:
	case 0xe7: return 65 | 0xff00; // F7
	case 0xd0:
	case 0xdc:
	case 0xe8: return 66 | 0xff00; // F8
	case 0xd1:
	case 0xdd:
	case 0xe9: return 67 | 0xff00; // F9
	case 0xd2:
	case 0xde:
	case 0xea: return 68 | 0xff00; // F10
	case 0xc0:
	case 0xf0: return 28 | 0xff00; // Enter

	case 0xed: return 15 | 0xff00; // TAB
	case 0xee: return 15 | 0x2a00; // s-TAB

	case 0x7f: return 14 | 0xff00; // BackSpace
	case 0xef: return 14 | 0x1d00; // c-BackSpace

	case 0xc1: return 75 | 0xe000; // left
	case 0xc5: return 75 | 0x2ae000;// s-left
	case 0xf1: return 75 | 0x1de000;// c-left

	case 0xc2: return 77 | 0xe000; // right
	case 0xc6: return 77 | 0x2ae000;// s-right
	case 0xf2: return 77 | 0x1de000;// c-right

	case 0xc3: return 80 | 0xe000; // down
	case 0xc7: return 80 | 0x2ae000;// s-down
	case 0xf3: return 80 | 0x1de000;// c-down

	case 0xc4: return 72 | 0xe000; // up
	case 0xc8: return 72 | 0x2ae000;// s-up
	case 0xf4: return 72 | 0x1de000;// c-down

	case 0xfc: return 58 | 0xff00; // capslock
//	case 0xfb: return 69 | 0xff00; // numlock

	case 0xf7: return 0xe051; // c-PgDn
	case 0xf8: return 0xe049; // c-PgUp

	case 0xfa: return 0xe052; // Ins
    }

    for(i = 0; i < 256; i++) {
	if (scan_unshifted[i] == asc) return i | 0xff00;
	if (scan_shift[i] == asc) return i | 0x2a00;
//	if (scan_unshifted[i] == asc) return i;
	if (scan_ctrl[i] == asc) return i | 0x1d00;
    }
    return 0xffff;
}

void get_keycode()
{
    static BYTE buffer[8];
    DWORD scank;
    BYTE key;
    int count;
    
    send_byte(READ_KEYCODE);
    count = 0;
    scank = asc2scan(recv_byte());
//    recv_byte();

    key = (scank >> 16) & 0xff;

    if (key != 0x0 || key !=0xff) {
        if (oldMkey != key){
	    if (oldMkey != 0xff || oldMkey != 0) buffer[count++] = oldMkey | 0x80;
	    buffer[count++] = key;
	    oldMkey = key;
	} else {
	}
    } else {
        if (oldMkey != 0xff || oldMkey != 0) {
	    buffer[count++] = oldMkey | 0x80;
	    oldMkey = 0xff;
	}
    }

    key = (scank >> 8) & 0xff;

    if (key != 0xff) {
        if (oldSkey != key){
	    if (oldSkey != 0xff) buffer[count++] = oldSkey | 0x80;
	    buffer[count++] = key;
	    oldSkey = key;
	} else {
	}
    } else {
        if (oldSkey != 0xff) {
	    buffer[count++] = oldSkey | 0x80;
	    oldSkey = 0xff;
	}
    }

    key = scank & 0xff;

    if (key != 0xff) {
        if (oldkey != key){
	    if (oldkey != 0xff) buffer[count++] = oldkey | 0x80;
	    buffer[count++] = key;
	    oldkey = key;
	} else {
	}
    } else {
        if (oldkey != 0xff) {
	    buffer[count++] = oldkey | 0x80;
	    oldkey = 0xff;
	}
    }
    	
    if (count) {
	printf("Key=%d\n", key);
        printf("oldKey=%d\n", oldkey);

        if (port60_buffer_ok(count)) {
	    printf("Buf ok!\n");
    	    put_scancode(buffer, count);
        }
    }
}
